@@ -31,14 +31,4 @@ __copyright__ = """
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 """
 
-import sys
-
-VERSION = sys.version_info
-PRE35 = VERSION.major >= 3 and VERSION.minor < 5
-
-# We need some pre-3.5 fixes, e.g. flatten() function.
-
-if PRE35:
-    from pyxcp.master.pre35 import Master
-else:
-    from pyxcp.master.py35 import Master
+from .master import Master
