@@ -12,7 +12,7 @@
 __copyright__ = """
     pySART - Simplified AUTOSAR-Toolkit for Python.
 
-   (C) 2009-2021 by Christoph Schueler <cpu12.gems@googlemail.com>
+   (C) 2009-2022 by Christoph Schueler <cpu12.gems@googlemail.com>
 
    All Rights Reserved
 
@@ -31,6 +31,7 @@ __copyright__ = """
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 """
 
+import functools
 import logging
 import struct
 import traceback
@@ -41,8 +42,9 @@ from pyxcp import checksum
 from pyxcp import types
 from pyxcp.config import Configuration
 from pyxcp.constants import makeBytePacker, makeByteUnpacker, makeWordPacker, makeWordUnpacker, makeDWordPacker, makeDWordUnpacker, makeDLongPacker, makeDLongUnpacker
-from pyxcp.master.errorhandler import wrapped
+from pyxcp.master.errorhandler import wrapped, disable_error_handling
 from pyxcp.transport.base import createTransport
+from pyxcp.utils import delay
 
 
 def broadcasted(func):
@@ -80,7 +82,10 @@ class Master:
 
     PARAMETER_MAP = {
         #            Type Req'd  Default
-        "LOGLEVEL": (str, False, "WARN"),
+        "LOGLEVEL":                         (str,  False, "WARN"),
+        "DISABLE_ERROR_HANDLING":           (bool, False, False),   # Bypass error-handling for performance reasons.
+        "SEED_N_KEY_DLL":                   (str,    False,  ""),
+        "SEED_N_KEY_DLL_SAME_BIT_WIDTH":    (bool, False, False),
     }
 
     def __init__(self, transportName, config=None):
@@ -89,6 +94,8 @@ class Master:
         self.config = Configuration(self.PARAMETER_MAP or {}, config or {})
         self.logger = logging.getLogger("pyXCP")
         self.logger.setLevel(self.config.get("LOGLEVEL"))
+        disable_error_handling(self.config.get("DISABLE_ERROR_HANDLING"))
+
         self.transport = createTransport(transportName, config)
         self.transport_name = transportName
 
@@ -112,7 +119,10 @@ class Master:
         self.mta = types.MtaType(None, None)
         self.currentDaqPtr = None
         self.currentProtectionStatus = None
-        self._seedNKeyDLL = None
+        self.seedNKeyDLL = self.config.get("SEED_N_KEY_DLL")
+        self.seedNKeyDLL_same_bit_width = self.config.get("SEED_N_KEY_DLL_SAME_BIT_WIDTH")
+        self.slaveProperties = SlaveProperties()
+        self.slaveProperties.pgmProcessor = SlaveProperties()
 
     def __enter__(self):
         """Context manager entry part."""
@@ -182,9 +192,10 @@ class Master:
         byteOrder = resultPartial.commModeBasic.byteOrder
 
         result = types.ConnectResponse.parse(response, byteOrder=byteOrder)
-        self.slaveProperties = SlaveProperties(byteOrder=byteOrder, maxCto=result.maxCto, maxDto=result.maxDto)
         byteOrderPrefix = "<" if byteOrder == types.ByteOrder.INTEL else ">"
-
+        self.slaveProperties.byteOrder = byteOrder
+        self.slaveProperties.maxCto = result.maxCto
+        self.slaveProperties.maxDto = result.maxDto
         self.slaveProperties.supportsPgm = result.resource.pgm
         self.slaveProperties.supportsStim = result.resource.stim
         self.slaveProperties.supportsDaq = result.resource.daq
@@ -197,7 +208,6 @@ class Master:
         self.slaveProperties.maxWriteDaqMultipleElements = (
             0 if self.slaveProperties.maxCto < 10 else int((self.slaveProperties.maxCto - 2) // 8)
         )
-
         self.BYTE_pack = makeBytePacker(byteOrderPrefix)
         self.BYTE_unpack = makeByteUnpacker(byteOrderPrefix)
         self.WORD_pack = makeWordPacker(byteOrderPrefix)
@@ -570,80 +580,160 @@ class Master:
 
     pull = fetch  # fetch() may be completely replaced by pull() someday.
 
-    def push(self, data: bytes):
+
+    def push(self, address: int, data: bytes, callback = None):
         """Convenience function for data-transfer from master to slave.
         (Not part of the XCP Specification).
 
+        Parameters
+        ----------
+        address: int
+
+        data : bytes
+            Arbitrary number of bytes.
+
+        Returns
+        -------
+        """
+        self._generalized_downloader(
+            address = address,
+            data = data,
+            maxCto = self.slaveProperties.maxCto,
+            maxBs = self.slaveProperties.maxBs,
+            minSt = self.slaveProperties.minSt,
+            master_block_mode = self.slaveProperties.masterBlockMode,
+            dl_func = self.download,
+            dl_next_func = self.downloadNext,
+            callback = callback
+        )
+
+    def flash_program(self, address: int, data: bytes, callback = None):
+        """Convenience function for flash programing.
+        (Not part of the XCP Specification).
+
+        Parameters
+        ----------
+        address: int
+
+        data : bytes
+            Arbitrary number of bytes.
+
+        Returns
+        -------
+        """
+        self._generalized_downloader(
+            address = address,
+            data = data,
+            maxCto = self.slaveProperties.pgmProcessor.maxCtoPgm,
+            maxBs = self.slaveProperties.pgmProcessor.maxBsPgm,
+            minSt = self.slaveProperties.pgmProcessor.minStPgm,
+            master_block_mode = self.slaveProperties.pgmProcessor.masterBlockMode,
+            dl_func = self.program,
+            dl_next_func = self.programNext,
+            callback = callback
+        )
+
+
+    def _generalized_downloader(
+            self,
+            address: int,
+            data: bytes,
+            maxCto: int,
+            maxBs: int,
+            minSt: int,
+            master_block_mode: bool,
+            dl_func,
+            dl_next_func,
+            callback = None
+        ):
+        """
+        """
+        self.setMta(address)
+        minSt /= 10000.0    # Unit is 100S.
+        block_downloader = functools.partial(
+            self._block_downloader,
+            dl_func = dl_func,
+            dl_next_func = dl_next_func,
+            minSt = minSt,
+        )
+        total_length = len(data)
+        if master_block_mode:
+            max_payload = min(maxBs * (maxCto - 2), 255)
+        else:
+            max_payload = maxCto - 2
+        offset = 0
+        if master_block_mode:
+            remaining = total_length
+            blocks = range(total_length // max_payload)
+            percent_complete = 1
+            remaining_block_size = total_length % max_payload
+            for idx in blocks:
+                block = data[offset : offset + max_payload]
+                block_downloader(block)
+                offset += max_payload
+                remaining -= max_payload
+                if callback and remaining <= total_length - (total_length / 100) * percent_complete:
+                    callback(percent_complete)
+                    percent_complete += 1
+            if remaining_block_size:
+                block = data[offset : offset + remaining_block_size]
+                block_downloader(block)
+                if callback:
+                    callback(percent_complete)
+        else:
+            chunk_size = max_payload
+            chunks = range(total_length // chunk_size)
+            remaining = total_length % chunk_size
+            for _ in chunks:
+                block = data[offset : offset + max_payload]
+                self.download(block, max_payload)
+                offset += max_payload
+            if remaining:
+                block = data[offset : offset + remaining]
+                self.download(block, remaining)
+
+
+    def _block_downloader(self, data: bytes, dl_func = None, dl_next_func = None, minSt = 0):
+        """Re-usable block downloader.
+
         Parameters
         ----------
         data : bytes
             Arbitrary number of bytes.
 
-        Returns
-        -------
+        dl_func: method
+            usually :meth: `download` or :meth:`program`
 
-        Note
-        ----
-        address is not included because of services implicitly setting address information like :meth:`getID` .
+        dl_next_func: method
+            usually :meth: `downloadNext` or :meth:`programNext`
+
+        minSt: int
+            Minimum separation time of frames.
         """
-        # TODO: consider minST.
-
-        total_length = len(data)
-        master_block_mode = self.slaveProperties.masterBlockMode
-        if master_block_mode:
-            max_payload = min(self.slaveProperties.maxBs * (self.slaveProperties.maxCto - 2), 255)
-        else:
-            max_payload = self.slaveProperties.maxCto - 2
-        offset = 0
-        if master_block_mode:
-            payload_length = self.slaveProperties.maxCto - 2
-            rem2 = total_length
-            blocks = range(total_length // max_payload)
-            remaining_block_size = total_length % max_payload
-            for idx in blocks:
-                data_slice = data[offset : offset + max_payload]
-                self._block_downloader(data_slice)
-                offset += max_payload
-                rem2 -= max_payload
-            if remaining_block_size:
-                data_slice = data[offset : offset + remaining_block_size]
-                self._block_downloader(data_slice)
-        else:
-            chunk_size = max_payload
-            chunks = range(total_length // chunk_size)
-            remaining = total_length % chunk_size
-            for _ in chunks:
-                frame_data = data[offset : offset + max_payload]
-                self.download(frame_data, max_payload)
-                offset += max_payload
-            if remaining:
-                frame_data = data[offset : offset + remaining]
-                self.download(frame_data, remaining)
-
-    def _block_downloader(self, data: bytes):
-        """"""
         length = len(data)
-        payload_length = self.slaveProperties.maxCto - 2
-        chunks = range(length // payload_length)
+        max_packet_size = self.slaveProperties.maxCto - 2  # Command ID + Length
+        packets = range(length // max_packet_size)
         offset = 0
-        remaining = length % payload_length
-        rem2 = length
-        idx = 0
-        for idx in chunks:
-            frame_data = data[offset : offset + payload_length]
-            if idx == 0:
-                self.download(frame_data, length)  # Transmit the complete length in the first CTO.
+        remaining = length % max_packet_size
+        remaining_block_size = length
+        index = 0
+        for index in packets:
+            packet_data = data[offset : offset + max_packet_size]
+            if index == 0:
+                dl_func(packet_data, length)  # Transmit the complete length in the first CTO.
             else:
-                self.downloadNext(frame_data, rem2)
-            offset += payload_length
-            rem2 -= payload_length
+                dl_next_func(packet_data, remaining_block_size)
+            offset += max_packet_size
+            remaining_block_size -= max_packet_size
+            delay(minSt)
         if remaining:
-            frame_data = data[offset : offset + remaining]
-            if idx == 0:
+            packet_data = data[offset : offset + remaining]
+            if index == 0:
                 # length of data is smaller than maxCto - 2
-                self.download(frame_data, remaining)
+                dl_func(packet_data, remaining, last = True)
             else:
-                self.downloadNext(frame_data, remaining)
+                dl_next_func(packet_data, remaining, last = True)
+            delay(minSt)
 
     @wrapped
     def download(self, data: bytes, blockModeLength=None):
@@ -1129,7 +1219,16 @@ class Master:
         `pyxcp.types.ProgramStartResponse`
         """
         response = self.transport.request(types.Command.PROGRAM_START)
-        return types.ProgramStartResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)
+        result = types.ProgramStartResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)
+        self.slaveProperties.pgmProcessor.commModePgm = result.commModePgm
+        self.slaveProperties.pgmProcessor.maxCtoPgm = result.maxCtoPgm
+        self.slaveProperties.pgmProcessor.maxBsPgm = result.maxBsPgm
+        self.slaveProperties.pgmProcessor.minStPgm = result.minStPgm
+        self.slaveProperties.pgmProcessor.queueSizePgm = result.queueSizePgm
+        self.slaveProperties.pgmProcessor.slaveBlockMode = result.commModePgm.slaveBlockMode
+        self.slaveProperties.pgmProcessor.interleavedMode = result.commModePgm.interleavedMode
+        self.slaveProperties.pgmProcessor.masterBlockMode = result.commModePgm.masterBlockMode
+        return result
 
     @wrapped
     def programClear(self, mode: int, clearRange: int):
@@ -1148,26 +1247,52 @@ class Master:
         return response
 
     @wrapped
-    def program(self, data: bytes):
-        """"""
-        d = bytearray()
-        d.append(len(data))
-        if self.slaveProperties.addressGranularity == types.AddressGranularity.DWORD:
-            d.extend(b"\x00\x00")  # alignment bytes
-        for e in data:
-            d.extend(self.AG_pack(e))
-        return self.transport.request(types.Command.PROGRAM, *d)
+    def program(self, data: bytes, blockLength, last=False):
+        """Parameters
+        ----------
+        data : bytes
+            Data to send to slave.
+        blockModeLength : int
+            the program request must contain the length of the whole block, not just the length
+            in the current packet.
+        last : bool
+            Indicates that this is the only packet in the block, because
+            the slave device will send the response after this.
+
+        Note
+        ----
+        Adress is set via :meth:`setMta`
+        """
+        #d = bytearray()
+        #d.append(len(data))
+        #if self.slaveProperties.addressGranularity == types.AddressGranularity.DWORD:
+        #    d.extend(b"\x00\x00")  # alignment bytes
+        #for e in data:
+        #    d.extend(self.AG_pack(e))
+        if last:
+            # last PROGRAM_NEXT packet in a block: the slave device has to send the response after this.
+            response = self.transport.request(types.Command.PROGRAM, blockLength, *data)
+            return response
+        else:
+            # the slave device won't respond to consecutive PROGRAM_NEXT packets in block mode,
+            # so we must not wait for any response
+            self.transport.block_request(types.Command.PROGRAM, blockLength, *data)
+            return None
+
 
     @wrapped
     def programReset(self):
         """Indicate the end of a programming sequence."""
-        return self.transport.request(types.Command.PROGRAM_RESET)
+        return self.transport.request_optional_response(types.Command.PROGRAM_RESET)
 
     @wrapped
     def getPgmProcessorInfo(self):
         """Get general information on PGM processor."""
         response = self.transport.request(types.Command.GET_PGM_PROCESSOR_INFO)
-        return types.GetPgmProcessorInfoResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)
+        result =  types.GetPgmProcessorInfoResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)
+        self.slaveProperties.pgmProcessor.pgmProperties = result.pgmProperties
+        self.slaveProperties.pgmProcessor.maxSector = result.maxSector
+        return result
 
     @wrapped
     def getSectorInfo(self, mode, sectorNumber):
@@ -1191,14 +1316,24 @@ class Master:
         )
 
     @wrapped
-    def programNext(self, data):
-        d = bytearray()
-        d.append(len(data))
-        if self.slaveProperties.addressGranularity == types.AddressGranularity.DWORD:
-            d.extend(b"\x00\x00")  # alignment bytes
-        for e in data:
-            d.extend(self.AG_pack(e))
-        return self.transport.request(types.Command.PROGRAM_NEXT, *d)
+    def programNext(self, data: bytes, remainingBlockLength: int,  last: bool = False):
+        #d = bytearray()
+        #d.append(len(data))
+        #if self.slaveProperties.addressGranularity == types.AddressGranularity.DWORD:
+        #    d.extend(b"\x00\x00")  # alignment bytes
+        #for e in data:
+        #    d.extend(self.AG_pack(e))
+        if last:
+            # last PROGRAM_NEXT packet in a block: the slave device has to send the response after this.
+            response = self.transport.request(types.Command.PROGRAM_NEXT, remainingBlockLength, *data)
+            return response
+        else:
+            # the slave device won't respond to consecutive PROGRAM_NEXT packets in block mode,
+            # so we must not wait for any response
+            self.transport.block_request(types.Command.PROGRAM_NEXT, remainingBlockLength, *data)
+            return None
+
+
 
     @wrapped
     def programMax(self, data):
@@ -1558,7 +1693,7 @@ class Master:
     def cond_unlock(self, resources=None):
         """Conditionally unlock resources, i.e. only unlock locked resources.
 
-        Precondition: Must assign :attr:`seedNKeyDLL`, e.g. ``master.seedNKeyDLL = "SeedNKeyXcp.dll"``
+        Precondition: Parameter "SEED_N_KEY_DLL" must be present and point to a valid DLL/SO.
 
         Parameters
         ----------
@@ -1582,7 +1717,7 @@ class Master:
 
         MAX_PAYLOAD = self.slaveProperties["maxCto"] - 2
 
-        if self._seedNKeyDLL is None:
+        if self.seedNKeyDLL is None:
             raise RuntimeError("No seed and key DLL specified, cannot proceed.")
         if resources is None:
             result = []
@@ -1614,7 +1749,7 @@ class Master:
                     result = self.getSeed(types.XcpGetSeedMode.REMAINING, resource_value)
                     seed.extend(list(result.seed))
                     remaining = result.length
-            result, key = getKey(self._seedNKeyDLL, resource_value, bytes(seed))
+            result, key = getKey(self.seedNKeyDLL, resource_value, bytes(seed), self.seedNKeyDLL_same_bit_width)
             if result == SeedNKeyResult.ACK:
                 key = list(key)
                 total_length = len(key)
@@ -1627,14 +1762,6 @@ class Master:
             else:
                 raise SeedNKeyError("SeedAndKey DLL returned: {}".format(SeedNKeyResult(result).name))
 
-    @property
-    def seedNKeyDLL(self):
-        return self._seedNKeyDLL
-
-    @seedNKeyDLL.setter
-    def seedNKeyDLL(self, name):
-        self._seedNKeyDLL = name
-
 def ticks_to_seconds(ticks, resolution):
     """Convert DAQ timestamp/tick value to seconds.
 
