@@ -4,7 +4,7 @@
 __copyright__ = """
     pySART - Simplified AUTOSAR-Toolkit for Python.
 
-   (C) 2009-2019 by Christoph Schueler <cpu12.gems@googlemail.com>
+   (C) 2009-2022 by Christoph Schueler <cpu12.gems@googlemail.com>
 
    All Rights Reserved
 
@@ -42,10 +42,13 @@ class Empty(Exception):
     pass
 
 
-def get(q, timeout):
+def get(q, timeout, restart_event):
     """Get an item from a deque considering a timeout condition."""
     start = time()
     while not q:
+        if restart_event.is_set():
+            start = time()
+            restart_event.clear()
         if time() - start > timeout:
             raise Empty
         sleep(0.001)
@@ -90,6 +93,7 @@ class BaseTransport(metaclass=abc.ABCMeta):
         timeout = self.config.get("TIMEOUT")
         self.alignment = self.config.get("ALIGNMENT")
         self.timeout = 2.0 if timeout is None else timeout
+        self.timer_restart_event = threading.Event()
         self.timing = Timing()
         self.resQueue = deque()
         self.daqQueue = deque()
@@ -146,18 +150,20 @@ class BaseTransport(metaclass=abc.ABCMeta):
         if hasattr(self, "closeEvent"):
             self.closeEvent.set()
 
-    def request(self, cmd, *data):
+    def _request_internal(self, cmd, ignore_timeout = False, *data):
         frame = self._prepare_request(cmd, *data)
         self.timing.start()
         self.send(frame)
 
         try:
-            xcpPDU = get(self.resQueue, timeout=self.timeout)
+            xcpPDU = get(self.resQueue, timeout=self.timeout, restart_event = self.timer_restart_event)
         except Empty:
-            raise types.XcpTimeoutError("Response timed out (timeout={}s)".format(self.timeout)) from None
-
+            if not ignore_timeout:
+                raise types.XcpTimeoutError("Response timed out (timeout={}s)".format(self.timeout)) from None
+            else:
+                self.timing.stop()
+                return
         self.timing.stop()
-
         pid = types.Response.parse(xcpPDU).type
         if pid == "ERR" and cmd.name != "SYNCH":
             err = types.XcpError.parse(xcpPDU[1:])
@@ -166,6 +172,12 @@ class BaseTransport(metaclass=abc.ABCMeta):
             pass  # Und nu??
         return xcpPDU[1:]
 
+    def request(self, cmd, *data):
+        return self._request_internal(cmd, False, *data)
+
+    def request_optional_response(self, cmd, *data):
+        return self._request_internal(cmd, True, *data)
+
     def block_request(self, cmd, *data):
         """
         Implements packet transmission for block communication model (e.g. DOWNLOAD block mode)
@@ -228,7 +240,6 @@ class BaseTransport(metaclass=abc.ABCMeta):
         ------
         :class:`pyxcp.types.XcpTimeoutError`
         """
-        TIMEOUT = 1.0  # TODO: parameter.
         block_response = b""
         start = time()
         while len(block_response) < length_required:
@@ -236,7 +247,7 @@ class BaseTransport(metaclass=abc.ABCMeta):
                 partial_response = self.resQueue.popleft()
                 block_response += partial_response[1:]
             else:
-                if time() - start > TIMEOUT:
+                if time() - start > self.timeout:
                     raise types.XcpTimeoutError("Response timed out [block_receive].") from None
                 sleep(0.001)
         return block_response
@@ -256,6 +267,13 @@ class BaseTransport(metaclass=abc.ABCMeta):
     def listen(self):
         pass
 
+    def process_event_packet(self, packet):
+        packet = packet[1 : ]
+        ev_type = packet[0]
+        self.logger.debug("EVENT-PACKET: {}".format(hexDump(packet)))
+        if ev_type == types.Event.EV_CMD_PENDING:
+            self.timer_restart_event.set()
+
     def processResponse(self, response, length, counter, recv_timestamp=None):
         if counter == self.counterReceived:
             self.logger.warn("Duplicate message counter {} received from the XCP slave".format(counter))
@@ -287,6 +305,7 @@ class BaseTransport(metaclass=abc.ABCMeta):
                 self.recv_timestamp = recv_timestamp
             elif pid == 0xFD:
                 # self.evQueue.put(response)
+                self.process_event_packet(response)
                 self.evQueue.append(response)
             elif pid == 0xFC:
                 # self.servQueue.put(response)
